@@ -3,6 +3,7 @@ use bevy_tasks::Task;
 use bevy_tasks::TaskPool;
 use bytes::Bytes;
 use std::{error::Error, net::SocketAddr, sync::{Arc, RwLock}};
+use instant::{Instant, Duration};
 
 use naia_client_socket::{
     ClientSocketTrait, MessageSender as ClientSender, Packet as ClientPacket,
@@ -32,21 +33,50 @@ pub type MultiplexedPacket = MuxPacket<<BufferPacketPool<SimpleBufferPool> as Pa
 pub type ConnectionChannelsBuilder =
     MessageChannelsBuilder<TaskPoolRuntime, MuxPacketPool<BufferPacketPool<SimpleBufferPool>>>;
 
-#[derive(Default, Debug, Clone)]
+#[derive(Debug, Clone)]
 pub struct PacketStats {
     pub packets_tx: usize,
     pub packets_rx: usize,
     pub bytes_tx: usize,
     pub bytes_rx: usize,
+    pub last_tx: Instant,
+    pub last_rx: Instant,
 }
+
+impl Default for PacketStats {
+    fn default() -> Self {
+        // default the last rx/tx to now.
+        // not strictly true in use-udp mode, since we can "connect" without
+        // exchanging packets. but always true for use-webrtc.
+        let now = Instant::now();
+        Self {
+            packets_tx: 0,
+            packets_rx: 0,
+            bytes_tx: 0,
+            bytes_rx: 0,
+            last_tx: now,
+            last_rx: now,
+         }
+    }
+}
+
 impl PacketStats {
     fn add_tx(&mut self, num_bytes: usize) {
         self.packets_tx += 1;
         self.bytes_tx += num_bytes;
+        self.last_tx = Instant::now();
     }
     fn add_rx(&mut self, num_bytes: usize) {
         self.packets_rx += 1;
         self.bytes_rx += num_bytes;
+        self.last_rx = Instant::now();
+    }
+    // returns Duration since last (rx, tx)
+    fn idle_durations(&self) -> (Duration, Duration) {
+        let now = Instant::now();
+        let rx = now.duration_since(self.last_rx);
+        let tx = now.duration_since(self.last_tx);
+        (rx, tx)
     }
 }
 
@@ -69,6 +99,9 @@ pub trait Connection: Send + Sync {
     fn channels_rx(&mut self) -> Option<&mut IncomingMultiplexedPackets<MultiplexedPacket>>;
 
     fn stats(&self) -> PacketStats;
+
+    /// returns milliseconds since last (rx, tx)
+    fn last_packet_timings(&self) -> (u128, u128);
 }
 
 #[cfg(not(target_arch = "wasm32"))]
@@ -82,6 +115,10 @@ pub struct ServerConnection {
 
     channels: Option<MessageChannels>,
     channels_rx: Option<IncomingMultiplexedPackets<MultiplexedPacket>>,
+    // channels_task is used to keep spawned Tasks in-scope so they are not dropped.
+    // we set it, but never read from it, which the compiler warns about.
+    // so we use allow(dead_code), even though it's needed for reference keeping.
+    #[allow(dead_code)]
     #[cfg(not(target_arch = "wasm32"))]
     channels_task: Option<Task<()>>,
 }
@@ -127,6 +164,11 @@ impl Connection for ServerConnection {
         )
     }
 
+    fn last_packet_timings(&self) -> (u128, u128) {
+        let (rx_dur, tx_dur) = self.stats.read().expect("stats lock poisoned").idle_durations();
+        (rx_dur.as_millis(), tx_dur.as_millis())
+    }
+
     fn receive(&mut self) -> Option<Result<Packet, NetworkError>> {
         match self.packet_rx.try_recv() {
             Ok(payload) => match payload {
@@ -192,6 +234,10 @@ pub struct ClientConnection {
 
     channels: Option<MessageChannels>,
     channels_rx: Option<IncomingMultiplexedPackets<MultiplexedPacket>>,
+    // channels_task is used to keep spawned Tasks in-scope so they are not dropped.
+    // we set it, but never read from it, which the compiler warns about.
+    // so we use allow(dead_code), even though it's needed for reference keeping.
+    #[allow(dead_code)]
     #[cfg(not(target_arch = "wasm32"))]
     channels_task: Option<Task<()>>,
 }
@@ -224,6 +270,11 @@ impl Connection for ClientConnection {
         self.stats.read().expect("stats lock poisoned").clone()
     }
 
+    fn last_packet_timings(&self) -> (u128, u128) {
+        let (rx_dur, tx_dur) = self.stats.read().expect("stats lock poisoned").idle_durations();
+        (rx_dur.as_millis(), tx_dur.as_millis())
+    }
+
     fn send(&mut self, payload: Packet) -> Result<(), Box<dyn Error + Sync + Send>> {
         self.stats.write().expect("stats lock poisoned").add_tx(payload.len());
         self.sender
